#!/usr/bin/env python3
#
#  apps.py
"""
Functions and classes for GitHub apps.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from datetime import date
from typing import Dict, Iterator, Optional

# 3rd party
import attr
from domdf_python_tools.stringlist import DelimitedList
from github3 import GitHub
from github3.apps import Installation
from typing_extensions import Literal

# this package
from github3_utils.headers import MACHINE_MAN

__all__ = ["ContextSwitcher", "iter_installed_repos", "make_footer_links"]


@attr.s
class ContextSwitcher:
	"""
	Class to aid switching contexts between the app itself and its installations.
	"""

	#:
	client: GitHub = attr.ib()

	#: The bytes of the private key for this GitHub App.
	private_key_pem: bytes = attr.ib()

	#: The integer identifier for this GitHub App.
	app_id: int = attr.ib()

	def login_as_app(self):
		"""
		Login as the GitHub app.
		"""

		self.client.login_as_app(self.private_key_pem, self.app_id)

	def login_as_user_installation(self, username: str) -> int:
		"""
		Login as a user installation of a GitHub app, and return its installation ID.

		:param username:
		"""

		# Log in as installation for this user
		installation_id = self.client.app_installation_for_user(username).id
		self.client.login_as_app_installation(self.private_key_pem, self.app_id, installation_id)

		return installation_id

	def login_as_org_installation(self, organization: str) -> int:
		"""
		Login as an organization installation of a GitHub app, and return its installation ID.

		.. versionadded:: 0.5.0

		:param organization:
		"""

		# Log in as installation for this org
		installation_id = self.client.app_installation_for_organization(organization).id
		self.client.login_as_app_installation(self.private_key_pem, self.app_id, installation_id)

		return installation_id

	def login_as_repo_installation(self, owner: str, repository: str) -> int:
		"""
		Login as a repository installation of a GitHub app, and return its installation ID.

		:param owner:
		:param repository:
		"""

		installation_id = self.client.app_installation_for_repository(
				owner=owner,
				repository=repository,
				).id
		self.client.login_as_app_installation(self.private_key_pem, self.app_id, installation_id)

		return installation_id


def iter_installed_repos(
		*,
		context_switcher: Optional[ContextSwitcher] = None,
		client: Optional[GitHub] = None,
		private_key_pem: Optional[bytes] = None,
		app_id: Optional[int] = None,
		) -> Iterator[Dict]:
	"""
	Returns an iterator over all repositories the app is installed for.

	:param context_switcher: A :class:`~.ContextSwitcher` used to switch contexts
		between the app itself and its installations.
	:param client: The bytes of the private key for this GitHub App.
	:param private_key_pem: The bytes of the private key for this GitHub App.
	:param app_id: The integer identifier for this GitHub App.

	Either ``context_switcher`` or all of ``client``, ``private_key_pem`` and ``app_id`` must be provided.
	"""

	if context_switcher is None:
		if client is None or private_key_pem is None or app_id is None:
			raise ValueError(
					"Either 'context_switcher' or all of 'client', "
					"'private_key_pem' and 'app_id' must be provided.",
					)
		else:
			context_switcher = ContextSwitcher(client, private_key_pem, app_id)

	context_switcher.login_as_app()
	client = context_switcher.client

	installation: Installation
	for installation in client.app_installations(context_switcher.app_id):
		# print(installation)
		# print(installation.account)
		username = installation.account["login"]

		context_switcher.login_as_app()
		context_switcher.login_as_user_installation(username)

		# Get repositories for this user.

		headers = {**installation.session.headers, **MACHINE_MAN}

		def get_page(page: int = 1):
			return context_switcher.client.session.get(  # type: ignore
				installation.repositories_url,
				params={"per_page": 100, "page": page},
				headers=headers,  # pylint: disable=cell-var-from-loop
				).json()

		response = get_page()
		total_repos = response["total_count"]
		yield from response["repositories"]

		total_repos -= len(response["repositories"])
		page = 2

		while total_repos > 0:
			response = get_page(page)
			page += 1
			yield from response["repositories"]
			total_repos -= len(response["repositories"])


_FooterType = Literal["marketplace", "app"]


def make_footer_links(
		owner: str,
		name: str,
		event_date: Optional[date] = None,
		type: _FooterType = "marketplace",  # noqa: A002  # pylint: disable=redefined-builtin
		docs_url: Optional[str] = None,
		) -> str:
	"""
	Create markdown footer links for a GitHub app.

	:param owner: The owner of the repository.
	:param name: The name of the repository.
	:param event_date: The date on which the footer is being created. Determines the emoji shown.
	:default event_data: today
	:param type: Whether the footer is for a GitHub app or an item in the marketplace.
	:param docs_url: The URL of the app's documentation. If :py:obj:`None` no link will be shown.

	.. versionadded:: 0.3.0
	"""

	if event_date is None:
		event_date = date.today()

	if event_date.month == 12 or (event_date.month == 1 and event_date.day <= 6):
		docs_emoji = '🎄'
		repo_emoji = '☃'
		issues_emoji = '🎅'
		marketplace_emoji = '🎁'
	else:
		docs_emoji = '📝'
		repo_emoji = ":octocat:"
		issues_emoji = '🙋'
		marketplace_emoji = '🏪'

	buf: DelimitedList[str] = DelimitedList()

	if docs_url:
		buf.append(f"[{docs_emoji} docs]({docs_url})")

	buf.extend([
			f"[{repo_emoji} repo](https://github.com/{owner}/{name})",
			f"[{issues_emoji} issues](https://github.com/{owner}/{name}/issues)",
			f"[{marketplace_emoji} marketplace](https://github.com/{type}/{name})",
			])

	return f"{buf: | }"



#!/usr/bin/env python3
#
#  __init__.py
"""
Handy utilities for `github3.py <https://github3py.readthedocs.io/en/master/>`_.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  "protect_branch" based on github3.py
#  Copyright (c) 2012-2016 by Ian Cordasco
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are
#  |  met:
#  |
#  |  1. Redistributions of source code must retain the above copyright
#  |  notice, this list of conditions and the following disclaimer.
#  |  2. Redistributions in binary form must reproduce the above copyright
#  |  notice, this list of conditions and the following disclaimer in the
#  |  documentation and/or other materials provided with the distribution.
#  |  3. The name of the author may not be used to endorse or promote products
#  |  derived from this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#  |  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
#  |  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  |  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  |  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  |  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#  |  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import datetime
import os
from contextlib import contextmanager
from typing import Iterable, Iterator, List, Optional, Union, overload

# 3rd party
import attr
from apeye import URL
from click import echo
from github3 import GitHub
from github3.orgs import Organization
from github3.repos import Repository, ShortRepository
from github3.repos.branch import Branch
from github3.users import User
from typing_extensions import Literal

# this package
from github3_utils.headers import LUKE_CAKE

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "0.5.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = [
		"RateLimitExceeded",
		"echo_rate_limit",
		"get_user",
		"protect_branch",
		"Impersonate",
		"get_repos",
		"iter_repos",
		]


class RateLimitExceeded(RuntimeError):
	"""
	Custom exception class to indicate the GitHub rate limit has been exceeded and no further requests should be made.
	"""

	#: The time at which the rate limit will be reset.
	reset_time: datetime.datetime

	def __init__(self, reset_time: datetime.datetime):
		super().__init__(f"No requests available! Resets at {reset_time}")
		self.reset_time = reset_time


@contextmanager
def echo_rate_limit(github: GitHub, verbose: bool = True):
	"""
	Contextmanager to echo the GitHub API rate limit before and after making a series of requests.

	:param github:
	:param verbose: If :py:obj:`False` no output will be printed.

	:raises: :exc:`click.Abort` if the rate limit has been exceeded.
	"""

	rate = github.rate_limit()["rate"]
	remaining_requests = rate["remaining"]
	reset = datetime.datetime.fromtimestamp(rate["reset"])

	if not remaining_requests:
		raise RateLimitExceeded(reset)

	if verbose:
		echo(f"{remaining_requests} requests available.")

	yield github

	if verbose:
		rate = github.rate_limit()["rate"]
		new_remaining_requests = rate["remaining"]
		used_requests = remaining_requests - new_remaining_requests
		reset = datetime.datetime.fromtimestamp(rate["reset"])

		echo(f"Used {used_requests} requests. {new_remaining_requests} remaining. Resets at {reset}")


def get_user(github: GitHub) -> User:
	"""
	Retrieve a :class:`github3.users.User` object for the authenticated user.

	:param github:
	"""

	url = github._build_url("user")
	json = github._json(github._get(url), 200)
	return github._instance_or_null(User, json)


def protect_branch(branch: Branch, status_checks: Optional[List[str]] = None) -> bool:
	"""
	Enable force push protection and configure status check enforcement.

	:param branch: The branch to enable protection for.
	:param status_checks: A list of strings naming status checks which must pass before merging.
		Use :py:obj:`None` or omit to use the already associated value.

	:returns: :py:obj:`True` if successful, :py:obj:`False` otherwise.
	"""

	previous_values = None
	previous_protection = getattr(branch, "original_protection", {})

	if previous_protection:
		previous_values = previous_protection.get("required_status_checks", {})

	if status_checks is None and previous_values:
		status_checks = previous_values["contexts"]

	edit = {
			"required_status_checks": {"strict": False, "contexts": status_checks},
			"enforce_admins": None,
			"required_pull_request_reviews": {
					"dismiss_stale_reviews": False,
					"required_approving_review_count": 1,
					},
			"restrictions": None,
			}

	resp = branch._put(
			str(URL(branch._api) / "protection"),
			json=edit,
			headers=LUKE_CAKE,
			)

	if branch._boolean(resp, 200, 404):
		branch.protected = True
		return True
	else:  # pragma: no cover
		return False


@attr.s
class Impersonate:
	"""
	Context manager to make commits as a specific user.

	Sets the following environment variables:

	* ``GIT_COMMITTER_NAME``
	* ``GIT_COMMITTER_EMAIL``
	* ``GIT_AUTHOR_NAME``
	* ``GIT_AUTHOR_EMAIL``

	.. attention::

		Any changes to environment variables made during the scope
		of the context manager will be reset on exit.

	**Example:**

	.. code-block:: python

		name = "repo-helper[bot]"
		email = f"74742576+{name}@users.noreply.github.com"

		commit_as_bot = Impersonate(name=name, email=email)

		with commit_as_bot():
			...

	"""

	#: The name of the committer.
	name: str = attr.ib()

	#: The email address of the committer.
	email: str = attr.ib()

	@contextmanager
	def __call__(self):
		"""
		The context manager itself.
		"""

		_environ = dict(os.environ)  # or os.environ.copy()

		try:
			name = "repo-helper[bot]"
			email = f"74742576+{name}@users.noreply.github.com"

			os.environ["GIT_COMMITTER_NAME"] = name
			os.environ["GIT_COMMITTER_EMAIL"] = email
			os.environ["GIT_AUTHOR_NAME"] = name
			os.environ["GIT_AUTHOR_EMAIL"] = email

			yield

		finally:
			os.environ.clear()
			os.environ.update(_environ)


@overload
def get_repos(
		user_or_org: Union[User, Organization],
		full: Literal[True],
		) -> Iterator[Repository]:
	...


@overload
def get_repos(
		user_or_org: Union[User, Organization],
		full: Literal[False] = ...,
		) -> Iterator[ShortRepository]:
	...


def get_repos(
		user_or_org: Union[User, Organization],
		full: bool = False,
		) -> Union[Iterator[Repository], Iterator[ShortRepository]]:
	"""
	Returns an iterator over the user or organisation's repositories.

	.. versionadded:: 0.5.0

	:param user_or_org:
	:param full: If :py:obj:`True` yields :class:`github3.repos.Repository` objects.
		Otherwise, yields :class:`github3.repos.ShortRepository` objects
	"""

	url = user_or_org._build_url("users", user_or_org.login, "repos")
	params = {"type": "owner", "sort": "full_name", "direction": "asc"}

	for repo in user_or_org._iter(-1, url, ShortRepository, params):
		if full:
			yield repo.refresh()
		else:
			yield repo


def iter_repos(
		github: GitHub,
		users: Iterable[str] = (),
		orgs: Iterable[str] = (),
		) -> Iterator[ShortRepository]:
	"""
	Returns an iterator over the repositories belonging to all ``users`` and all ``orgs``.

	.. versionadded:: 0.5.0

	:param github:
	:param users: An iterable of usernames to fetch the repositories for.
	:param orgs: An iterable of organization names to fetch the repositories for.
	"""

	for user in users:
		_user: Optional[User] = github.user(user)
		if _user is None:
			raise ValueError(f"No such user {user}")

		yield from get_repos(_user, full=False)

	for org in orgs:
		_org: Optional[Organization] = github.organization(org)

		if _org is None:
			raise ValueError(f"No such organization {org}")

		yield from get_repos(_org, full=False)



#!/usr/bin/env python3
#
#  check_labels.py
"""
Helpers for creating labels to mark pull requests with which tests are failing.

.. versionadded:: 0.4.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from typing import Dict, NamedTuple, Set, Union

# 3rd party
import attr
import github3.issues.label
from domdf_python_tools.doctools import prettify_docstrings
from github3.checks import CheckRun
from github3.issues import Issue
from github3.pulls import PullRequest, ShortPullRequest
from github3.repos import Repository
from github3.repos.commit import ShortCommit

__all__ = ["Label", "check_status_labels", "Checks", "get_checks_for_pr", "label_pr_failures"]


@prettify_docstrings
@attr.s(frozen=True, slots=True)
class Label:
	"""
	Represents an issue or pull request label.
	"""

	#: The text of the label.
	name: str = attr.ib(converter=str)

	#: The background colour of the label.
	color: str = attr.ib(converter=str)

	#: A short description of the label.
	description: str = attr.ib(default=None)

	def __str__(self) -> str:
		return self.name

	def to_dict(self) -> Dict[str, str]:
		"""
		Return the :class:`~.Label` as a dictionary.
		"""

		return {
				"name": self.name,
				"color": self.color,
				"description": self.description,
				}

	def create(self, repo: Repository) -> github3.issues.label.Label:
		"""
		Create this label on the given repository.

		:param repo:
		"""

		return repo.create_label(**self.to_dict())


check_status_labels: Dict[str, Label] = {
		label.name: label
		for label in [
				Label("failure: flake8", "#B60205", "The Flake8 check is failing."),
				Label("failure: mypy", "#DC1C13", "The mypy check is failing."),
				Label("failure: docs", "#EA4C46", "The docs check is failing."),
				Label("failure: Windows", "#F07470", "The Windows tests are failing."),
				Label("failure: Linux", "#F6BDC0", "The Linux tests are failing."),
				# Label("failure: Multiple", "#D93F0B", "Multiple checks are failing."),
				]
		}
"""
Labels corresponding to failing pull request checks.
"""

# The ``failure: Multiple`` label is used if three or more categories failing.


class Checks(NamedTuple):
	"""
	Represents the sets of status checks returned by :func:`~.get_checks_for_pr`.
	"""

	successful: Set[str]
	failing: Set[str]
	running: Set[str]
	skipped: Set[str]
	neutral: Set[str]


def get_checks_for_pr(pull: Union[PullRequest, ShortPullRequest]) -> Checks:
	"""
	Returns a :class:`~.Checks` object containing sets of check names grouped by their status.

	:param pull: The pull request to obtain checks for.
	"""

	head_commit: ShortCommit = list(pull.commits())[-1]

	failing = set()
	running = set()
	successful = set()
	skipped = set()
	neutral = set()

	check_run: CheckRun
	for check_run in head_commit.check_runs():

		if check_run.status in {"queued", "running", "in_progress"}:
			running.add(check_run.name)
		elif check_run.conclusion in {"failure", "cancelled", "timed_out", "action_required"}:
			failing.add(check_run.name)
		elif check_run.conclusion == "success":
			successful.add(check_run.name)
		elif check_run.conclusion == "skipped":
			skipped.add(check_run.name)
		elif check_run.conclusion == "neutral":
			neutral.add(check_run.name)

	# Remove failing checks from successful etc. (as all checks appear twice for PRs)
	successful = successful - failing - running
	running = running - failing
	skipped = skipped - running - failing - successful
	neutral = neutral - running - failing - successful

	return Checks(
			successful=successful,
			failing=failing,
			running=running,
			skipped=skipped,
			neutral=neutral,
			)


_python_dev_re = re.compile(r".*Python\s*\d+\.\d+.*(dev|alpha|beta|rc).*", flags=re.IGNORECASE)


def label_pr_failures(pull: Union[PullRequest, ShortPullRequest]) -> Set[str]:
	"""
	Labels the given pull request to indicate which checks are failing.

	:param pull:

	:return: The new labels set for the pull request.
	"""

	pr_checks = get_checks_for_pr(pull)

	failure_labels: Set[str] = set()
	success_labels: Set[str] = set()

	def determine_labels(from_, to):
		for check in from_:
			if _python_dev_re.match(check):
				continue

			if check in {"Flake8", "docs"}:
				to.add(f"failure: {check.lower()}")
			elif check.startswith("mypy"):
				to.add("failure: mypy")
			elif check.startswith("ubuntu"):
				to.add("failure: Linux")
			elif check.startswith("windows"):
				to.add("failure: Windows")

	determine_labels(pr_checks.failing, failure_labels)
	determine_labels(pr_checks.successful, success_labels)

	issue: Issue = pull.issue()

	current_labels = {label.name for label in issue.labels()}

	for label in success_labels:
		if label in current_labels and label not in failure_labels:
			issue.remove_label(label)

	current_labels -= success_labels
	current_labels.update(failure_labels)

	issue.add_labels(*current_labels)

	return current_labels



#!/usr/bin/env python3
#
#  click.py
"""
Extensions for `click <https://click.palletsprojects.com/en/7.x/>`_.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Callable, TypeVar

# 3rd party
import click

__all__ = ["token_option"]

_C = TypeVar("_C", bound=click.Command)


def token_option(token_var: str = "GITHUB_TOKEN") -> Callable[[_C], _C]:  # nosec: B107
	"""
	Creates a ``-t / --token`` option for the GitHub API token.

	.. versionadded:: 0.2.0

	:param token_var:
	"""

	return click.option(
			"-t",
			"--token",
			type=click.STRING,
			help=(
					"The token to authenticate with the GitHub API. "
					f"Can also be provided via the '{token_var}' environment variable."
					),
			envvar=token_var,
			required=True,
			)



# stdlib
import sys

__all__ = ["make_typed_dict"]

if sys.version_info > (3, 9):  # pragma: no cover (<py39)
	# stdlib
	from typing import TypedDict

	make_typed_dict = TypedDict

else:  # pragma: no cover (py39+)
	# 3rd party
	from typing_extensions import _TypedDictMeta

	def make_typed_dict(typename, *args, total=True, **kwargs):  # pragma: no cover
		"""
		Create a TypedDict.

		Temporary fix until https://github.com/python/typing/issues/761 is merged.

		:param typename:
		:param args:
		:param total:
		:param kwargs:
		"""

		# From CPython
		# PSF Licensed
		# Copyright Python Software Foundation

		if args:
			try:
				fields, = args  # allow the "_fields" keyword be passed
			except ValueError:
				raise TypeError(
						f'TypedDict.__new__() takes from 2 to 3 positional arguments but {len(args) + 2} were given'
						)
		else:
			raise TypeError("TypedDict.__new__(): not enough arguments")

		if fields is None:
			fields = kwargs
		elif kwargs:
			raise TypeError("TypedDict takes either a dict or keyword arguments, but not both")

		ns = {"__annotations__": dict(fields), "__total__": total}
		try:
			# Setting correct module is necessary to make typed dict classes pickleable.
			ns["__module__"] = sys._getframe(1).f_globals.get("__name__", "__main__")
		except (AttributeError, ValueError):
			pass

		return _TypedDictMeta(typename, (), ns, total=total)



#!/usr/bin/env python3
#
#  testing.py
"""
Fixtures for `pytest <https://docs.pytest.org/en/stable/>`_.

.. extras-require:: testing
	:__pkginfo__:

.. versionadded:: 0.2.0

To use this module you need to add, at a minimum, the following to your ``conftest.py``:

.. code-block:: python

	from betamax import Betamax

	pytest_plugins = ("github3_utils.testing", )

	with Betamax.configure() as config:
		config.cassette_library_dir = "<path to cassettes directory>"
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# 3rd party
import pytest  # nodep
from _pytest.fixtures import FixtureRequest  # nodep
from betamax import Betamax  # type: ignore  # nodep
from github3 import GitHub

__all__ = ["cassette", "github_client", "module_cassette"]


@pytest.fixture()
def github_client() -> GitHub:
	"""
	Provides an instance of :class:`github3.github.GitHub`,
	using a fake token to authenticate.
	"""  # noqa: D400

	return GitHub(token="FAKE_TOKEN")  # nosec: B106


@pytest.fixture()
def cassette(request: FixtureRequest, github_client):
	"""
	Provides a Betamax cassette scoped to the test function
	which record and plays back interactions with the GitHub API.
	"""  # noqa: D400

	with Betamax(github_client.session) as vcr:
		vcr.use_cassette(request.node.name, record="none")

		yield github_client


@pytest.fixture()
def module_cassette(request: FixtureRequest, github_client):
	"""
	Provides a Betamax cassette scoped to the test module
	which record and plays back interactions with the GitHub API.
	"""  # noqa: D400

	cassette_name = request.module.__name__.split('.')[-1]

	with Betamax(github_client.session) as vcr:
		# print(f"Using cassette {cassette_name!r}")
		vcr.use_cassette(cassette_name, record="none")

		yield github_client
